#!/usr/bin/env python

import re
import tkinter
import pdb
from collections import defaultdict
import numpy as np

def get_dic(st):
    """
    assemble string into dict
    :param st: string
    :return: dictionary
    """
    assert len(st) % 2 == 0, 'array order is not right'
    dc = {}
    for i in range(len(st) // 2):
        a = st[i * 2]
        b = st[i * 2 + 1]

        # convert dict value to float if possible
        try:
            val = float(b)
        except ValueError:
            val = b
        dc[a] = val

    return dc


def read_array(r, name, subdic=True):
    """
    read tcl array
    :param r: tkinter object
    :param name: array name
    :param subdic: group array in sub-dicts if possible?
    :return: dictionary
    """
    exe = r.tk.eval('array get ' + name)

    # extract dict components
    st = list(filter(None, re.split(' |({|})', exe)))

    # replace empty dicts by -1
    st_clean = []
    i = 0
    while i < len(st) - 1:
        if st[i] == '{' and st[i + 1] == '}':
            st_clean.append('-1')
            i += 1
        else:
            st_clean.append(st[i])
        i += 1
    if i == len(st) - 1:
        st_clean.append(st[-1])

    # check for sub-dicts
    limits = [i for i, s in enumerate(st_clean) if s == '{' or s == '}']
    assert len(limits) % 2 == 0, 'array order is not right'

    # one dictionary
    if len(limits) == 0:
        dc = get_dic(st_clean)

    # several sub-dictionaries
    else:
        dc = {}
        for a, b in zip(limits[::2], limits[1::2]):
            if not subdic:
                dc[st_clean[a - 1]] = st_clean[a + 1:b]
            else:
                dc[st_clean[a - 1]] = get_dic(st_clean[a + 1:b])
    return dc


def get_in_model_units(s_units, symbol, val):
    # equal units, no conversion
    if s_units == 'cm':
        return val

    # convert units
    else:
        sign = +1.0
        # if s_units == 'mm' and m_units == 'cm':
        #     sign = +1.0
        # elif s_units == 'cm' and m_units == 'mm':
        #     sign = -1.0
        # else:
        #     raise ValueError('Unknown unit combination ' + s_units + ' and ' + m_units)

        if symbol == 'R' or symbol == 'q':
            return val * np.power(10.0, sign * 4)
        elif symbol == 'C':
            return val * np.power(10.0, - sign * 4)
        elif symbol == 'P':
            return val * np.power(10.0, sign * 1)
        elif 'density' in symbol:
            return val * np.power(10.0, sign * 3)
        elif 'viscosity' in symbol:
            return val * np.power(10.0, sign * 1)
        else:
            raise ValueError('Unknown boundary condition symbol ' + symbol)


def get_flow_coronary(r):
    # read tcl array
    exe = r.tk.eval('array get sim_array')
    st = list(filter(None, re.split(' |({|})', exe)))
    st = [i for i in st if i != '{' and i != '}']

    # create new dict key for each pressure table
    pressures = defaultdict(list)
    name = ''
    for e in st:
        try:
            val = float(e)
            pressures[name] += [val]
        except ValueError:
            name = e

    # split arrays in time and pressure
    for k, v in pressures.items():
        pressures[k] = np.array(v).reshape(-1, 2)

    return pressures


def get_bcs(tcl, tcl_bc):
    # evaluate tcl-script to extract variables for boundary conditions
    r_bc = tkinter.Tk()
    r_bc.tk.eval('source ' + tcl_bc)

    # generate dictionaries from tcl output
    sim_bc = read_array(r_bc, 'sim_bc')
    sim_spid = read_array(r_bc, 'sim_spid')
    sim_preid = read_array(r_bc, 'sim_preid')
    sim_spname = read_array(r_bc, 'sim_spname', False)

    # get parameters
    params = get_params(tcl)

    # get boundary condition types
    sim_bc_type = get_bc_type(sim_bc)

    # convert boundary condition parameters to sim units
    for cp, bc in sim_bc.items():
        if is_outlet(cp):
            # add zero reference pressure if not defined
            if sim_bc_type[cp] in ('rcr', 'resistance') and 'Po' not in bc:
                sim_bc[cp]['Po'] = 0.0
            for p, v in bc.items():
                if p[0] in ('R', 'C'):
                    sim_bc[cp][p] = get_in_model_units(params['sim_units'], p[0], v)

    # extract pressure over time in case of coronary boundary conditions
    coronary = {}
    for bc in sim_bc.values():
        if 'COR' in bc.keys():
            coronary = get_flow_coronary(r_bc)
            for cp, v in coronary.items():
                coronary[cp][:, 1] = get_in_model_units(params['sim_units'], 'P', v[:, 1])

    # convert model parameters to sim units
    for p, v in params.items():
        if p in ('sim_density', 'sim_viscosity'):
            params[p] = get_in_model_units(params['sim_units'], p, float(v))

    bcs = {'bc': sim_bc, 'spid': sim_spid, 'preid': sim_preid, 'spname': sim_spname, 'coronary': coronary,
           'params': params, 'bc_type': sim_bc_type}

    # close window
    r_bc.destroy()

    return bcs


def get_bc_type(bc_def):
    bc_type = {}
    for s, bc in bc_def.items():
        if not is_outlet(s):
            continue
        if 'Rp' in bc and 'C' in bc and 'Rd' in bc:
            bc_type[s] = 'rcr'
        elif 'R' in bc and 'Po' in bc:
            bc_type[s] = 'resistance'
        elif 'COR' in bc:
            bc_type[s] = 'coronary'
        else:
            raise RuntimeError('boundary conditions not implemented')

    return bc_type


def is_outlet(s):
    if 'wall' in s or 'inflow' in s or 'stent' in s:
        return False
    else:
        return True


def get_params(tcl):
    # evaluate tcl-script to extract variables for general simulation parameters
    r = tkinter.Tk()
    r.tk.eval('source ' + tcl)
    params = {}
    for v in r.tk.eval('info globals').split():
        try:
            var = r.tk.getvar(v)

            # don't add tcl stuff to dict
            if isinstance(var, str) or isinstance(var, int) or isinstance(var, float):
                params[v] = var
            elif isinstance(var, tuple):
                params[v] = list(var)
            else:
                try:
                    params[v] = var.string
                except:
                    pass
        except:
            pass

    # close window
    r.destroy()

    return params



#!/usr/bin/env python
import vtk
import os
import meshio
import argparse
import glob
import numpy as np
import pdb
from collections import defaultdict

from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from get_bc_integrals import get_res_names, read_geo, write_geo
from get_database import Database, input_args
from vtk_functions import read_geo, write_geo, cell_connectivity, collect_arrays


def jacobian_positive(points, tets):
    """
    Check if Jacobian determinant is positive everywhere
    """
    jac = np.zeros((tets.shape[0], 3, 3))
    for i, p in enumerate([[1, 0], [2, 0], [3, 0]]):
        jac[:, :, i] = points[tets[:, p[0]]] - points[tets[:, p[1]]]
    return np.sum(np.linalg.det(jac) <= 0.0) == 0


def get_last_timestep(res, field):
    """
    Return the last time step of a field
    """
    return [k for k in sorted(res.keys()) if field in k][-1]


def get_last_result(fpath):
    """
    Return the results of the last time step for pressure and velocity
    """
    if not os.path.exists(fpath):
        raise ValueError('No results found in ' + fpath)
    geo = read_geo(fpath).GetOutput()
    res = collect_arrays(geo.GetPointData())
    return res[get_last_timestep(res, 'pressure')], res[get_last_timestep(res, 'velocity')]


def get_initial_conditions(db, geo, point_data, ini):
    """
    Generate initial conditions from database for pressure and velocity
    """
    # load initial conditions
    ini_val = defaultdict(dict)

    # zero initial conditions (as in SimVascular)
    if 'zero' in ini.values():
        n_points = len(point_data['GlobalNodeID'])
        ini_val['pressure']['zero'] = np.zeros(n_points)
        ini_val['velocity']['zero'] = 0.0001 * np.ones((n_points, 3))

    # initial conditions from file
    ini_paths = {'steady': db.get_initial_conditions_steady(geo),
                 'steady0': db.get_initial_conditions_steady0(geo),
                 'irene': db.get_initial_conditions_irene(geo),
                 'osmsc': db.get_volume(geo),
                 'asymp': db.get_asymptotic(geo)}

    for i, f_path in ini_paths.items():
        if i in ini.values():
            ini_val['pressure'][i], ini_val['velocity'][i] = get_last_result(f_path)

    fpath_red = db.get_initial_conditions_pressure(geo)
    if '1d' in ini.values() and os.path.exists(fpath_red):
        data_1d = read_geo(fpath_red).GetOutput().GetPointData()
        for f in ini.keys():
            if data_1d.HasArray(f):
                ini_val[f]['1d'] = v2n(data_1d.GetArray(f))

    # apply initial conditions
    for f, i in ini.items():
        if i not in ini_val[f]:
            raise RuntimeError('Unknown ' + f + ' initialization ' + i)

        print('  initial condition ' + f + ': ' + i)
        point_data[f] = ini_val[f][i]


def get_vol(db, geo, ini):
    """
    Generate volume mesh for SimVascular: remove all unused arrays and reoder tet nodes
    """
    f_vol = db.get_volume(geo)
    f_out = os.path.join(db.get_sv_meshes(geo), geo + '.vtu')
    f_ini = db.get_initial_conditions(geo)

    if not os.path.exists(f_vol):
        print('  no volume mesh')
        return

    # read volume mesh
    vol = read_geo(f_vol).GetOutput()

    # get geometry
    points = v2n(vol.GetPoints().GetData())
    cells = cell_connectivity(vol)

    # reorder nodes in tets to fix negative Jacobian
    if not jacobian_positive(points, cells['tetra']):
        cells['tetra'] = cells['tetra'][:, [0, 1, 3, 2]]
        print('  tets flipped')
    else:
        print('  tets ok')
    # assert jacobian_positive(points, cells['tetra']), 'Jacobian negative after flipping tets'

    if not jacobian_positive(points, cells['tetra']):
        print('  Jacobian negative after flipping tets')

    # get arrays
    point_data = {'GlobalNodeID': np.expand_dims(v2n(vol.GetPointData().GetArray('GlobalNodeID')), axis=1)}
    cell_data = {'GlobalElementID': np.expand_dims(v2n(vol.GetCellData().GetArray('GlobalElementID')), axis=1)}

    # write raw write to file
    mesh = meshio.Mesh(points, [('tetra', cells['tetra'])], point_data=point_data, cell_data=cell_data)
    meshio.write(f_out, mesh)

    # get initial conditions
    get_initial_conditions(db, geo, point_data, ini)

    # write initial conditions to file
    mesh = meshio.Mesh(points, [('tetra', cells['tetra'])], point_data=point_data, cell_data=cell_data)
    meshio.write(f_ini, mesh)


def get_indices(a, b):
    """
    Elementwise True if an entry in a is in any of b
    """
    i = np.zeros(a.shape, dtype=bool)
    for j in b:
        i |= (a == j)
    return i


def get_surf(db, geo):
    """
    Generate arrays in surface mesh used by SimVascular
    """
    # get all surfaces
    surfaces = glob.glob(os.path.join(db.get_surface_dir(geo), '*.vtp'))

    for f_surf in surfaces:
        # read volume mesh with results
        surf = read_geo(f_surf).GetOutput()
        surf_p = surf.GetPointData()
        surf_c = surf.GetCellData()

        # get output name
        name_osmsc = os.path.basename(f_surf)
        if 'all_exterior' in name_osmsc:
            name = geo + '.vtp'
        elif 'wall' in name_osmsc:
            name = 'walls_combined.vtp'
        else:
            name = os.path.join('caps', name_osmsc)

        f_out = os.path.join(db.get_sv_meshes(geo), name)

        # reconstruct SimVascular arrays from BC_FaceID
        face_id = v2n(surf_c.GetArray('BC_FaceID'))

        # read surface ids
        try:
            caps = get_indices(face_id, db.get_surface_ids(geo, 'caps'))
            inflow = get_indices(face_id, db.get_surface_ids(geo, 'inflow'))
            outlets = get_indices(face_id, db.get_surface_ids(geo, 'outlets'))
        except (KeyError, TypeError):
            print('  face missing in boundary conditions')
            return

        # initialize new arrays
        n_names = ['GlobalBoundaryPoints']
        c_names = ['GlobalBoundaryCells', 'CapID', 'BadTriangle', 'FreeEdge', 'BooleanRegion', 'ModelFaceID',
                   'Normals', 'ActiveCells']
        arrays = {}
        for n in n_names:
            arrays[n] = {'handle': surf_p, 'array': np.zeros(surf.GetNumberOfPoints(), dtype=np.int64)}
        for n in c_names:
            arrays[n] = {'handle': surf_c, 'array': np.zeros(surf.GetNumberOfCells(), dtype=np.int64)}

        # rename
        arrays['ModelFaceID']['array'] = face_id + 1

        # all caps
        arrays['ActiveCells']['array'][caps] = 1

        # inflow is 1, outflow is 2
        arrays['CapID']['array'][inflow] = 1
        arrays['CapID']['array'][outlets] = 2

        # remove old array
        surf_c.RemoveArray('BC_FaceID')

        # add new arrays
        for n, v in arrays.items():
            out_array = n2v(v['array'])
            out_array.SetName(n)
            v['handle'].AddArray(out_array)

        # generate normals
        normals = vtk.vtkPolyDataNormals()
        normals.SetInputData(surf)
        normals.ComputePointNormalsOff()
        normals.ComputeCellNormalsOn()
        normals.SplittingOff()
        normals.Update()

        # export to generated folder
        write_geo(f_out, normals.GetOutput())


def get_meshes(db, geo):
    ini = {}
    if 'ini_' in db.study:
        ini_type = db.study.split('_')[1]
        ini['pressure'] = ini_type
        ini['velocity'] = ini_type
    elif 'steady' in db.study:
        ini['pressure'] = 'zero'
        ini['velocity'] = 'zero'
    elif 'irene' == db.study:
        ini['pressure'] = 'steady'
        ini['velocity'] = 'steady'
    else:
        raise ValueError('Unknown intialization for study ' + db.study)

    get_vol(db, geo, ini)
    get_surf(db, geo)


def main(db, geometries):
    """
    Loop all geometries
    """
    for geo in geometries:
        print('Running geometry ' + geo)

        get_meshes(db, geo)


if __name__ == '__main__':
    descr = 'Generate all meshes for SimVascular'
    d, g, _ = input_args(descr)
    main(d, g)



# Optimization for Vessel Connections

import numpy as np
from geomdl import BSpline, utilities
from geomdl.visualization import VisMPL
import numba as nb
from scipy import interpolate, optimize
import pyvista as pv

@nb.jit(nopython=True)
def get_angle(V1,V2):
    return np.arccos(np.dot(-V1,V2))*(180/np.pi)

@nb.jit(nopython=True)
def get_all_angles(vectors):
    angles = np.zeros(vectors.shape[0]-1)
    for i in range(vectors.shape[0]-1):
        angles[i] = get_angle(vectors[i,:],vectors[i+1,:])
    return angles

@nb.jit(nopython=True)
def close_exact(data,point,radius_buffer):
    line_direction = np.zeros((data.shape[0],3))
    ss = np.zeros(data.shape[0])
    tt = np.zeros(data.shape[0])
    hh = np.zeros(data.shape[0])
    cc = np.zeros((data.shape[0],3))
    cd = np.zeros(data.shape[0])
    line_distances = np.zeros(data.shape[0])
    for i in range(data.shape[0]):
        line_direction[i,:] = (data[i,3:6] - data[i,0:3])/np.linalg.norm(data[i,3:6] - data[i,0:3])
        ss[i] = np.dot(data[i,0:3]-point,line_direction[i,:])
        tt[i] = np.dot(point-data[i,3:6],line_direction[i,:])
        d = np.array([ss[i],tt[i],0])
        hh[i] = np.max(d)
        diff = point - data[i,0:3]
        cc[i,:] = np.cross(diff,line_direction[i,:])
        cd[i] = np.linalg.norm(cc[i,:])
        line_distances[i] = np.sqrt(hh[i]**2+cd[i]**2) - data[i,6] - radius_buffer
    return line_distances

def cost_angles(angles):
    #func = lambda x: np.tanh(-(x+np.arctanh(np.log(2))))+np.log(1+np.exp(-x))
    func = lambda x: np.log(1+np.exp(-x*100))
    return np.sum(func(angles))

def get_all_vectors(pts):
    vectors = pts[1:] - pts[:-1]
    vectors = vectors/np.linalg.norm(vectors,axis=1).reshape(-1,1)
    return vectors

def check_bounds(bounds,sample_pts):
    bounds_score = 0
    #func = lambda x: np.tanh((x-np.arctanh(np.log(2))))+np.log(1+np.exp(x))
    #func = lambda x: np.exp(x)
    func = lambda x: np.log(1+np.exp(x*100))
    for i in range(sample_pts.shape[0]):
        bounds_score += func(bounds[0,0] - sample_pts[i,0])
        bounds_score += func(sample_pts[i,0] - bounds[0,1])
        bounds_score += func(bounds[1,0] - sample_pts[i,1])
        bounds_score += func(sample_pts[i,1] - bounds[1,1])
        bounds_score += func(bounds[2,0] - sample_pts[i,2])
        bounds_score += func(sample_pts[i,2] - bounds[2,1])
    return bounds_score

def get_collisions(collision_vessels,R,sample_pts,radius_buffer):
    collisions = 0
    #func = lambda x: np.tanh((x-np.arctanh(np.log(2))))+np.log(1+np.exp(x))
    #func = lambda x: np.log(1+np.exp(x))

    # Check for collisions with other vessels
    func = lambda x: 0.5*np.tanh(x*100)+0.5
    for i in range(sample_pts.shape[0]):
        dist = close_exact(collision_vessels,sample_pts[i,:],radius_buffer)
        collisions += np.sum(func(R + radius_buffer - dist))
    # Check for collisions for vessel with itself
    #same_vessel = np.zeros((sample_pts.shape[0]-1,7))
    #same_vessel[:,0:3] = sample_pts[:-1,:]
    #same_vessel[:,3:6] = sample_pts[1:,:]
    #same_vessel[:,6]   = R
    #mid_points = (sample_pts[:-1,:] - sample_pts[1:,:])/2
    #for i in range(mid_points.shape[0]):
    #    if i <= 
    #    subset_same_vessels = 
    return collisions

def connect_bezier(P1,P2,P3,P4,clamp_first=True,clamp_second=True,number_vessels=20):
    if clamp_first and clamp_second:
        V1   = (P2 - P1)
        L1   = np.linalg.norm(V1)
        V1   = V1/np.linalg.norm(V1)
        V2   = (P4 - P3)
        L2   = np.linalg.norm(V2)
        V2   = V2/np.linalg.norm(V2)
        def create_bezier(data,V1=V1,V2=V2,P2=P2,P4=P4,L1=L1,L2=L2):
            CTR0 = P2 + (L1/2+data[2])*V1
            CTR1 = P4 + (L2/2+data[3])*V2
            CTR  = np.zeros(data.shape[0]-4+12)
            fill_CTR = np.zeros(data.shape[0]-4).reshape(-1,3)
            seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
            for i in range(fill_CTR.shape[0]):
                fill_CTR[i,:] = CTR0*(1-seg[i]) + CTR1*(seg[i])
            fill_CTR = fill_CTR.flatten()
            CTR[0:3]     = P1 + (L1/4+data[0])*V1
            CTR[3:6]     = CTR0
            CTR[-6:-3]   = CTR1
            CTR[-3:]     = P3 + (L2/4+data[1])*V2
            CTR[6:-6]    += fill_CTR + data[4:]
            CTR  = CTR.reshape(-1,3).tolist()
            curve = BSpline.Curve()
            curve.degree = len(CTR) - 1
            curve.ctrlpts = CTR
            curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
            curve.sample_size = number_vessels
            curve.evaluate()
            return curve
    elif not clamp_first and clamp_second:
        V2   = (P4 - P3)
        L2   = np.linalg.norm(V2)
        V2   = V2/np.linalg.norm(V2)
        def create_bezier(data,V2=V2,P2=P2,P4=P4,L2=L2):
            CTR1 = P4 + (L2/2+data[1])*V2
            CTR  = np.zeros(data.shape[0]-2+9)
            fill_CTR = np.zeros(data.shape[0]-2).reshape(-1,3)
            seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
            for i in range(fill_CTR.shape[0]):
                fill_CTR[i,:] = P2*(1-seg[i]) + CTR1*(seg[i])
            fill_CTR = fill_CTR.flatten()
            CTR[0:3]     = P2
            CTR[-6:-3]   = CTR1
            CTR[-3:]     = P3 + (L2/4+data[0])*V2
            CTR[3:-6]    += fill_CTR + data[2:]
            CTR  = CTR.reshape(-1,3).tolist()
            curve = BSpline.Curve()
            curve.degree = len(CTR) - 1
            curve.ctrlpts = CTR
            curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
            curve.sample_size = 20
            curve.evaluate()
            return curve
    elif clamp_first and not clamp_second:
        V1   = (P2 - P1)
        L1   = np.linalg.norm(V1)
        V1   = V1/np.linalg.norm(V1)
        def create_bezier(data,V1=V1,P2=P2,P4=P4,L1=L1):
            CTR0 = P2 + (L1/2+data[1])*V1
            CTR  = np.zeros(data.shape[0]-2+9)
            fill_CTR = np.zeros(data.shape[0]-2).reshape(-1,3)
            seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
            for i in range(fill_CTR.shape[0]):
                fill_CTR[i,:] = CTR0*(1-seg[i]) + P4*(seg[i])
            fill_CTR = fill_CTR.flatten()
            CTR[0:3]     = P1 + (L1/4+data[0])*V1
            CTR[3:6]     = CTR0
            CTR[-3:]     = P4
            CTR[6:-3]    += fill_CTR + data[2:]
            CTR  = CTR.reshape(-1,3).tolist()
            curve = BSpline.Curve()
            curve.degree = len(CTR) - 1
            curve.ctrlpts = CTR
            curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
            curve.sample_size = 20
            curve.evaluate()
            return curve
    else:
        def create_bezier(data,P2=P2,P4=P4):
            CTR  = np.zeros(data.shape[0]+6)
            fill_CTR = np.zeros(data.shape[0]).reshape(-1,3)
            seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
            for i in range(fill_CTR.shape[0]):
                fill_CTR[i,:] = P2*(1-seg[i]) + P4*(seg[i])
            fill_CTR = fill_CTR.flatten()
            CTR[0:3]     = P2
            CTR[-3:]     = P4
            CTR[3:-3]    += fill_CTR + data
            CTR  = CTR.reshape(-1,3).tolist()
            curve = BSpline.Curve()
            curve.degree = len(CTR) - 1
            curve.ctrlpts = CTR
            curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
            curve.sample_size = 20
            curve.evaluate()
            return curve
    return create_bezier

def bezier_cost(data,create_curve=None,R=None,
                P1=None,P3=None,collision_vessels=None,
                radius_buffer=0,bounds=None,sample_size=20,
                clamp_first=True,clamp_second=True):
    curve = create_curve(data)
    curve.sample_size = sample_size
    curve.evaluate()
    pts   = np.array(curve.evalpts)
    sample_pts = np.array(curve.evalpts)
    if collision_vessels is not None:
        collisions = get_collisions(collision_vessels,R,sample_pts,radius_buffer)
    else:
        collisions = 0
    bound_score = check_bounds(bounds,sample_pts)
    if clamp_first and clamp_second:
        pts = np.vstack((P1,pts,P3))
    elif clamp_first and not clamp_second:
        pts = np.vstack((P1,pts))
    elif not clamp_first and clamp_second:
        pts = np.vstack((pts,P3))
    else:
        pass
    vectors = get_all_vectors(pts)
    angles  = get_all_angles(vectors)
    angles  = cost_angles(angles-110)
    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    connection_length_min = np.linalg.norm(P3-P1)
    spline_length_norm = spline_length/connection_length_min
    return (bound_score+angles+collisions)*spline_length_norm

class connection:
    def __init__(self,forest,network,tree_idx,tree_jdx,terminal_index,radius_buffer,
                 clamp_first=True,clamp_second=True,seperate_connection=True,
                 number_vessels=20):
        self.network  = network
        self.tree_idx = tree_idx
        self.tree_jdx = tree_jdx
        self.edge_idx = forest.assignments[self.network][self.tree_idx][terminal_index]
        self.edge_jdx = forest.assignments[self.network][self.tree_jdx][terminal_index]
        self.edge_1   = forest.networks[self.network][self.tree_idx].data[self.edge_idx,:]
        self.edge_2   = forest.networks[self.network][self.tree_jdx].data[self.edge_jdx,:]
        self.tree_idx_last_index = int(np.max(forest.networks[self.network][self.tree_idx].data[:,-1]))+1
        self.tree_jdx_last_index = int(np.max(forest.networks[self.network][self.tree_jdx].data[:,-1]))+1
        self.P1       = self.edge_1[0:3]
        self.P2       = self.edge_1[3:6]
        self.P3       = self.edge_2[0:3]
        self.P4       = self.edge_2[3:6]
        self.R1       = self.edge_1[21]
        self.R2       = self.edge_2[21]
        self.bounds   = np.array([forest.boundary.x_range,forest.boundary.y_range,forest.boundary.z_range])
        self.best     = np.inf
        self.xopt     = None
        self.number_vessels = number_vessels
        self.generator = connect_bezier(self.P1,self.P2,self.P3,self.P4,clamp_first=clamp_first,clamp_second=clamp_second,number_vessels=number_vessels)
        self.boundary = forest.boundary.pv_polydata
        tree_idx_data_indicies = np.argwhere(forest.networks[self.network][self.tree_idx].data[:,15]>-1).flatten()
        tree_jdx_data_indicies = np.argwhere(forest.networks[self.network][self.tree_jdx].data[:,15]>-1).flatten()
        tree_idx_data = forest.networks[self.network][self.tree_idx].data[tree_idx_data_indicies,:]
        tree_jdx_data = forest.networks[self.network][self.tree_jdx].data[tree_jdx_data_indicies,:]
        self.upstream_tree_idx = [self.edge_1]
        self.upstream_tree_jdx = [self.edge_2]
        self.collision_vessels = np.zeros((tree_idx_data.shape[0]+tree_jdx_data.shape[0],7))
        self.collision_vessels[:tree_idx_data.shape[0],0:3]  = tree_idx_data[:,0:3]
        self.collision_vessels[tree_idx_data.shape[0]:,0:3]  = tree_jdx_data[:,0:3]
        self.collision_vessels[:tree_idx_data.shape[0],3:6]  = tree_idx_data[:,3:6]
        self.collision_vessels[tree_idx_data.shape[0]:,3:6]  = tree_jdx_data[:,3:6]
        self.collision_vessels[:tree_idx_data.shape[0],6]    = tree_idx_data[:,21]
        self.collision_vessels[tree_idx_data.shape[0]:,6]    = tree_jdx_data[:,21]
        self.radius_buffer = radius_buffer
        self.clamp_first = clamp_first
        self.clamp_second = clamp_second
        self.seperate_connection = seperate_connection
    def solve(self,number_free_points=2,iters=1,maxtime=None,f_min=None):
        self.create_curve = connect_bezier(self.P1,self.P2,self.P3,self.P4,
                                           clamp_first=self.clamp_first,
                                           clamp_second=self.clamp_second,
                                           number_vessels=self.number_vessels)
        cost = lambda data: bezier_cost(data,create_curve=self.create_curve,
                                        R=self.R1,P1=self.P1,P3=self.P3,
                                        collision_vessels=self.collision_vessels,
                                        sample_size=self.number_vessels,
                                        radius_buffer=self.radius_buffer,bounds=self.bounds,
                                        clamp_first=self.clamp_first,clamp_second=self.clamp_second)
        if self.clamp_first and self.clamp_second:
            clamp_number = 4
        elif self.clamp_first or self.clamp_second:
            clamp_number = 2
        else:
            clamp_number = 0
        if self.xopt is None:
            self.xopt = np.zeros(clamp_number+3*number_free_points)
        if not self.xopt.shape[0] == clamp_number+3*number_free_points:
            self.xopt = np.zeros(clamp_number+3*number_free_points)
        L = np.linalg.norm(self.P2-self.P4)
        lb = np.ones(self.xopt.shape[0])*(-L)
        for i in range(clamp_number):
            lb[i] = 0
        ub = np.ones(self.xopt.shape[0])*L
        ctrl_bounds = []
        for b in range(len(lb)):
            ctrl_bounds.append([lb[b],ub[b]])
        if number_free_points == 1:
            if maxtime is None and f_min is None:
                res = optimize.shgo(cost,bounds=ctrl_bounds,iters=iters)
            elif not maxtime is None:
                res = optimize.shgo(cost,bounds=ctrl_bounds,iters=iters,options={'maxtime':maxtime})
            elif not f_min is None:
                res = optimize.shgo(cost,bounds=ctrl_bounds,iters=iters,options={'f_min':f_min})
        else:
            res = optimize.basinhopping(cost,self.xopt,niter=iters)
        self.xopt = res.x
        self.best = res.fun
        return res
    def check_angles(self):
        curve = self.create_curve(self.xopt)
        curve.sample_size = self.number_vessels
        curve.evaluate()
        pts   = np.array(curve.evalpts)
        pts = np.vstack((self.P1,pts,self.P3))
        vectors = get_all_vectors(pts)
        angles  = get_all_angles(vectors)
        return cost_angles(angles)
    def check_bounds(self):
        curve = self.create_curve(self.xopt)
        curve.sample_size = self.number_vessels
        curve.evaluate()
        pts   = np.array(curve.evalpts)
        return check_bounds(self.bounds,pts)
    def check_collisions(self):
        curve = self.create_curve(self.xopt)
        curve.sample_size = self.number_vessels
        curve.evaluate()
        pts   = np.array(curve.evalpts)
        return get_collisions(self.collision_vessels,max(self.R1,self.R2),pts,self.radius_buffer)
    def show(self):
        tree_1_vessels = []
        tree_2_vessels = []
        connection_vessels = []
        other_vessels = []
        """
        for vessel in range(len(self.upstream_tree_idx)):
            center    = (self.upstream_tree_idx[vessel][0:3] + self.upstream_tree_idx[vessel][3:6])/2
            direction = self.upstream_tree_idx[vessel][3:6] - self.upstream_tree_idx[vessel][0:3]
            direction = direction/np.linalg.norm(direction)
            radius    = self.upstream_tree_idx[vessel][21]
            length    = self.upstream_tree_idx[vessel][20]
            cylinder  = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            tree_1_vessels.append(cylinder)
        for vessel in range(len(self.upstream_tree_jdx)):
            center    = (self.upstream_tree_jdx[vessel][0:3] + self.upstream_tree_jdx[vessel][3:6])/2
            direction = self.upstream_tree_jdx[vessel][3:6] - self.upstream_tree_jdx[vessel][0:3]
            direction = direction/np.linalg.norm(direction)
            radius    = self.upstream_tree_jdx[vessel][21]
            length    = self.upstream_tree_jdx[vessel][20]
            cylinder  = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            tree_2_vessels.append(cylinder)
        """
        plotter = pv.Plotter()
        if not self.vessels is None:
            colors = ['r','b']
            for i in range(len(self.vessels)):
                for j in range(self.vessels[i].shape[0]):
                    center = (self.vessels[i][j,0:3] + self.vessels[i][j,3:6])/2
                    cylinder = pv.Cylinder(center=center,direction=self.vessels[i][j,12:15],radius=self.vessels[i][j,21],height=self.vessels[i][j,20])
                    plotter.add_mesh(cylinder,color=colors[i])
        for vessel in range(self.collision_vessels.shape[0]):
            center = (self.collision_vessels[vessel,0:3] + self.collision_vessels[vessel,3:6])/2
            direction = self.collision_vessels[vessel,3:6] - self.collision_vessels[vessel,0:3]
            length = np.linalg.norm(direction)
            direction = direction/length
            radius = self.collision_vessels[vessel,-1]
            cylinder  = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            plotter.add_mesh(cylinder,color='black')
        plotter.add_mesh(self.boundary,opacity=0.25)
        plotter.show()
    def build_vessels(self):
        curve = self.create_curve(self.xopt)
        curve.sample_size = self.number_vessels
        curve.evaluate()
        pts     = np.array(curve.evalpts)
        pts     = np.vstack((self.P1,pts,self.P3))
        if self.seperate_connection:
            sep     = (pts.shape[0]-1)//2
            vessels_tree_idx = np.ones((sep,self.edge_1.shape[0]))*-1
            vessels_tree_jdx = np.ones(((pts.shape[0]-1)-sep,self.edge_1.shape[0]))*-1
            tree_idx_pts     = pts[:sep+1,:]
            tree_jdx_pts     = np.flip(pts[sep:,:],axis=0)
            vessels_tree_idx[:,0:3] = tree_idx_pts[:-1,:]
            vessels_tree_idx[:,3:6] = tree_idx_pts[1:,:]
            vessels_tree_jdx[:,0:3] = tree_jdx_pts[:-1,:]
            vessels_tree_jdx[:,3:6] = tree_jdx_pts[1:,:]
            tree_idx_directions     = (tree_idx_pts[1:] - tree_idx_pts[:-1])
            tree_jdx_directions     = (tree_jdx_pts[1:] - tree_jdx_pts[:-1])
            vessels_tree_idx[:,20]  = np.linalg.norm(tree_idx_directions,axis=1)
            vessels_tree_jdx[:,20]  = np.linalg.norm(tree_jdx_directions,axis=1)
            vessels_tree_idx[:,12:15] = tree_idx_directions/np.linalg.norm(tree_idx_directions,axis=1).reshape(-1,1)
            vessels_tree_jdx[:,12:15] = tree_jdx_directions/np.linalg.norm(tree_jdx_directions,axis=1).reshape(-1,1)
            vessels_tree_idx[:,21]  = self.R1
            vessels_tree_jdx[:,21]  = self.R2
            vessels_tree_idx[:,22]  = self.edge_1[22]
            vessels_tree_jdx[:,22]  = self.edge_2[22]
            self.vessels = [vessels_tree_idx,vessels_tree_jdx]
        else:
            vessels = np.ones((pts.shape[0]-1,self.edge_1.shape[0]))*-1
            vessels[:,0:3]    = pts[:-1,:]
            vessels[:,3:6]    = pts[1:,:]
            vessels[:,21]     = max(self.R1,self.R2)
            vessel_directions = pts[1:] - pts[:-1]
            vessels[:,20]     = np.linalg.norm(vessel_directions,axis=1)
            vessels[:,12:15]  = vessel_directions/np.linalg.norm(vessel_directions,axis=1).reshape(-1,1)
            vessels[:,22]     = self.edge_1[22]
            self.vessels = [vessels]
        return
    def add_collision_vessels(self,connection_object):
        connection_object.build_vessels()
        for vessel_group in connection_object.vessels:
            tmp = np.zeros((vessel_group.shape[0],7))
            tmp[:,0:6] = vessel_group[:,0:6]
            tmp[:,6] = vessel_group[:,21]
            self.collision_vessels = np.vstack((self.collision_vessels,tmp))
    def replace_terminals(self,forest):
        pass



#!/usr/bin/env python

import paraview.simple as pv
import numpy as np
import os
import pdb

from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from common import input_args
from get_bc_integrals import read_geo, write_geo
from get_database import Database


def write_geo_pv(fname_in, fname_out, ele_id, aname):
    """
    Read/write geometry and remove/add cell array.
    Not sure how replacing array works with ParaView reader, so read again here using vtk
    :param fname_in: input filename
    :param fname_out: output filename
    :param ele_id: cell array to be written
    :param aname: cell array name
    :return: void
    """
    out_array = n2v(ele_id)
    out_array.SetName(aname)
    reader, _, reader_cell = read_geo(fname_in)
    reader_cell.RemoveArray(aname)
    reader_cell.AddArray(out_array)
    write_geo(fname_out, reader)


def read_geo_pv(fname):
    """
    Read geometry using built-in ParaView functions. Distinguish between 2D/3D meshes.
    In case of 3D mesh, extract surface
    :param fname: geometry file name
    :return: reader (for writer), reader (to extract data), point data, cell data
    """
    _, ext = os.path.splitext(fname)
    if ext == '.vtp':
        reader = pv.XMLPolyDataReader(FileName=[fname])
    elif ext == '.vtu':
        reader_3d = pv.XMLUnstructuredGridReader(FileName=[fname])
        reader = pv.ExtractSurface(Input=reader_3d)
    else:
        raise ValueError('File extension ' + ext + ' unknown.')
    reader_fetch = pv.servermanager.Fetch(reader)
    return reader, reader_fetch, reader_fetch.GetPointData(), reader_fetch.GetCellData()


def is_unique(ids):
    """
    Check if list of ids is unqiue
    :param ids: 1d numpy array of ids
    :return: True if ids is unique
    """
    return np.unique(ids).shape[0] == ids.shape[0]


def read_array(data, name):
    """
    Read vtk array from data and convert to numpy int array
    :param data: vtk cell/point data
    :param name: name of array
    :return: 1-d numpy int array
    """
    return v2n(data.GetArray(name)).astype(int)


def get_connectivity(reader):
    """
    Extract surface cell connectivity from mesh
    :param reader: vtk reader
    :return: [N,3] numpy array with N surface elements in mesh containing point ids
    """
    # loop all cells
    cells = []
    for i in range(reader.GetNumberOfCells()):
        c = reader.GetCell(i)

        # loop all points of current cell
        points = []
        for j in range(c.GetNumberOfPoints()):
            points.append(c.GetPointIds().GetId(j))

        # extract only surface elements
        if len(points) == 3:
            cells.append(points)
    return np.array(cells)


def get_ids(fpath):
    """
    Read geometry and extract relevant node/element ids
    :param fpath: filename
    :return: vtk reader,  GlobalElementID, element connectivity in GlobalNodeID (points sorted for each element)
    """
    reader, reader_fetch, point_data, cell_data = read_geo_pv(fpath)

    node_id = read_array(point_data, 'GlobalNodeID')
    assert is_unique(node_id), 'GlobalNodeID is not unique'

    cell_id = read_array(cell_data, 'GlobalElementID')
    connectivity = get_connectivity(reader_fetch)

    return reader, cell_id, np.sort(node_id[connectivity], axis=1)


def fix_surfaces(fpath_vol, fpath_surf, folder_out):
    """
    Problem: GlobalElementID in 2D meshes only stored with 6-digit precision
    Solution: Compare elements in 2D and 3D meshes to restore GlobalElementID
    :param fpath_vol: filename of 3D mesh
    :param fpath_surf: filenames of corresponding 2D meshes
    :param folder_out: output folder
    :return: True of all 2D meshes were converted successfully
    """
    # read volume mesh
    vol_reader, vol_cell, vol_conn = get_ids(fpath_vol)

    # loop all corresponding surface meshes
    for f in fpath_surf:
        surf_fname = os.path.basename(f)
        if not surf_fname == 'wall.vtp':
            continue
        print('  mesh ' + surf_fname)

        # get surface mesh
        surf_reader, surf_cell, surf_conn = get_ids(f)

        # skipping (GlobalElementID already unique)
        if is_unique(surf_cell):
            surf_cell_new = surf_cell

        # match GlobalElementID in surface mesh with volume mesh
        else:
            # todo: this can probably be done without loops
            surf_cell_new = np.zeros(surf_cell.shape, dtype=int)
            for i, cell in enumerate(surf_conn):
                found = vol_cell[(vol_conn == cell).all(axis=1)]

                # check if element was found in both 2d and 3d mesh
                if not found.shape[0] == 1:
                    print('    error: 2d and 3d mesh are not identical')
                    return False

                surf_cell_new[i] = found[0]

            # compare original 2D GlobalElementID to the fixed one
            assert np.max(np.abs(surf_cell_new - surf_cell)) <= 5, 'round-off error bigger than 5'

        # export surface
        fpath_out = os.path.join(folder_out, surf_fname)
        write_geo_pv(f, fpath_out, surf_cell_new, 'GlobalElementID')
        del surf_reader
    else:
        del vol_reader
        return True


def is_fixed(folder_out, fpath_surf):
    """
    Check if all surfaces meshes have been fixed for a geometry
    :param folder_out: output folder for fixed meshes
    :param fpath_surf: list of surface mesh files
    :return: True if ALL surface meshes have been fixed
    """
    for f in fpath_surf:
        if not os.path.exists(os.path.join(folder_out, os.path.basename(f))):
            return False
    else:
        return True


def main(db, geometries):
    # loop all geometries in repository
    database_surf_fixed = {}
    for geo in geometries:
        print('Fixing geometry ' + geo)
        folder_out = os.path.join(db.fpath_gen, 'surfaces', geo)
        try:
            os.mkdir(folder_out)
        except OSError:
            pass

        # get volume mesh path
        fpath_vol = db.get_volume(geo)

        # get all surface mesh paths
        fpath_surf = db.get_surfaces_upload(geo)

        if is_fixed(folder_out, fpath_surf):
            print('  skipping (already fixed)')
            database_surf_fixed[geo] = True
        elif not os.path.exists(fpath_vol):
            print('  skipping (no 3D mesh found)')
            database_surf_fixed[geo] = False
        else:
            database_surf_fixed[geo] = fix_surfaces(fpath_vol, fpath_surf, folder_out)

    fpath_report = os.path.join(db.fpath_gen, 'database', 'surf_fixed')
    np.save(fpath_report, database_surf_fixed)


if __name__ == '__main__':
    descr = 'Fix wrong GlobalElementID'
    d, g, _ = input_args(descr)
    main(d, g)


#!/usr/bin/env python

import contextlib
import csv
import glob
import io
import os
import pdb
import re
import shutil
import sys
import argparse
import subprocess

import numpy as np

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

from get_database import Database, SimVascular, Post, input_args
from simulation_io import get_dict
from vtk_functions import read_geo, write_geo, collect_arrays

import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from matplotlib.offsetbox import OffsetImage, AnnotationBbox


def plot_error_spatial(db, geometries):
    # set global plot options
    fig, ax = plt.subplots(2, 1, figsize=(8, 8), dpi=200, sharey='all')#, sharex='all'
    plt.rcParams['axes.linewidth'] = 2

    # read errors
    f_path = db.get_3d_3d_comparison()
    if not os.path.exists(f_path):
        return
    err = get_dict(f_path)

    # select geometries
    geos = [geo for geo in geometries if geo in err]
    legend = [geo[:4] for geo in geometries if geo in err]

    for i, f in enumerate(['pressure', 'velocity']):
        for j, c in enumerate(['avg']):#, 'max'
            # plot location
            pos = i#(i, j)

            # plot data points
            lengths = []
            for geo in geos:
                e = err[geo][f][c]
                x = np.arange(1, len(e) + 1)
                lengths += [len(e)]
                ax[pos].plot(x, e, 'o-')

            # set plot options
            ax[pos].grid(True)
            ax[pos].set_xlabel('Cardiac cycle')
            if f == 'pressure':
                err_name = r'$\epsilon_{p,avg}$'
            elif f == 'velocity':
                err_name = r'$\epsilon_{u,avg}$'
            ax[pos].set_ylabel(f.capitalize() + ' relative ' + c + '. error ' + err_name)
            ax[pos].ticklabel_format(axis='y') #
            # r"$\epsilon_{p,\text{avg}$"
            # ax[pos].legend(legend)
            ax[pos].set_yscale('log')
            ax[pos].set_xticks(np.arange(1, np.max(lengths) + 1))
            ax[pos].yaxis.set_major_formatter(mtick.PercentFormatter(1.0))

    lgd = ax[(0)].legend(legend, bbox_to_anchor=(-0.2, 0), loc='right') #
    # fig.subplots_adjust(left=0.2)
    # plt.subplots_adjust(left=0.07, right=0.93, wspace=0.25, hspace=0.35)
    plt.subplots_adjust(right=0.8)
    fname = '3d_3d_comparison.png'
    fpath = os.path.join(db.get_statistics_dir(), fname)
    fig.savefig(fpath, bbox_extra_artists=(lgd,), bbox_inches='tight')#
    plt.close(fig)


def main(db, geometries, params):
    plot_error_spatial(db, geometries)


if __name__ == '__main__':
    descr = 'Get 3D-3D statistics'
    d, g, p = input_args(descr)
    main(d, g, p)



#!/usr/bin/env python

import contextlib
import csv
import glob
import io
import os
import pdb
import re
import shutil
import sys
import argparse
import subprocess

import numpy as np
from collections import defaultdict

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

from get_database import Database, SimVascular, Post, input_args
from simulation_io import get_dict, get_caps_db, collect_results_db_3d_3d
from vtk_functions import read_geo, write_geo, collect_arrays
from compare_1d import plot_1d_3d_caps

import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from matplotlib.offsetbox import OffsetImage, AnnotationBbox


def plot_error_spatial(db, geometries):
    # get post-processing constants
    post = Post()
    fields = post.fields
    fields.remove('area')

    # set global plot options
    fig, ax = plt.subplots(len(fields), len(geometries), figsize=(16, 8), dpi=200, sharey=True, sharex=True)  #
    plt.rcParams['axes.linewidth'] = 2

    # get 1d/3d map
    cycles = []
    for j, geo in enumerate(geometries):
        # get results
        res, time = collect_results_db_3d_3d(db, geo)

        # number of cardiac cycles
        n_cycle = time['3d_rerun_n_cycle']
        cycles += [n_cycle]

        # get caps
        caps = get_caps_db(db, geo)

        for i, f in enumerate(post.fields):
            # plot location
            if len(geometries) == 1:
                pos = i
            else:
                pos = (i, j)

            # calculate errors
            err = []
            err_last = []
            val = []
            for c, br in caps.items():
                # get results for branch at all time steps
                res_br = res[br][f]['3d_rerun_all']

                # get last cycle
                res_last = res[br][f]['3d_rerun_cap']

                # normalize
                if f == 'pressure':
                    norm = np.mean(res_last)
                elif f == 'flow':
                    norm = np.max(res_last) - np.min(res_last)

                # get start and end step of each cardiac cycle
                cycle_range = []
                for k in range(1, n_cycle + 1):
                    i_cycle = np.where(time['3d_rerun_i_cycle_' + str(k)])[0]
                    cycle_range += [i_cycle]

                # calculate cycle error
                err_br = []
                err_br_last = []
                val_br = []
                t0 = []
                for k in range(1, n_cycle):
                    t_prev = cycle_range[k - 1]
                    t_this = cycle_range[k]
                    t_last = cycle_range[-1]
                    diff = np.mean(res_br[t_this] - res_br[t_prev])
                    err_br += [np.abs(diff / norm)]
                    diff = np.mean(res_br[t_last] - res_br[t_prev])
                    err_br_last += [np.abs(diff / norm)]
                    val_br += [res_br[t_this[0]]]
                    t0 += [time['3d_rerun_all'][t_prev[0]]]
                err += [err_br]
                err_last += [err_br_last]
                val += [val_br]
                t0 = np.array(t0)
            err = np.array(err).T
            err_last = np.array(err_last).T
            val = np.array(val)

            # mean error over all outlets
            err_all = np.mean(err, axis=1)

            # how much does the error decrease in each cycle?
            slope = np.mean(np.diff(np.log10(err_all)))

            # error threshold for a converged solution
            thresh = 1.0e-3

            # how many cycles are needed to reach convergence
            i_conv = np.where(np.all(err < thresh, axis=1))[0]
            if not i_conv.any():
                i_conv = n_cycle + int((np.log10(thresh) - np.log10(err_all[-1])) / slope + 1.0)
            else:
                i_conv = i_conv[0] + 2

            # time constants per outlet [s]
            if f == 'pressure':
                dt = np.mean(np.diff(t0))
                dval_dt = np.diff(val) / np.diff(t0)

                # select all time steps where error decreases
                i_decr = np.where(np.all(dval_dt > 0, axis=0) == False)[0]
                if np.any(i_decr):
                    dval_dt = dval_dt[:, :i_decr[0]]
                    t0 = t0[:i_decr[0]]
                else:
                    t0 = t0[:-1]

                tau_all = - 1 / np.mean(np.diff(np.log(dval_dt)) / dt, axis=1)
                val0 = np.zeros(len(caps))
                for cap, tau in enumerate(tau_all):
                    val0[cap] += [np.mean(tau * dval_dt[cap] / np.exp(-t0 / tau))]
                out_str = geo + ' pres time constant ' + '{:2.1e}'.format(np.max(tau_all)) + ' [s]'
                out_str += ' = ' '{:2.1e}'.format(np.max(tau_all) / dt) + ' [cycles]'
                print(out_str)

            # calculate periodic value
            f_conv = norm
            for m in range(10000 - n_cycle):
                f_conv += err_all[-1] * norm * 10 ** ((m + 1) * slope)

            # print errors
            max_err = np.max(err[-1])
            max_outlet = db.get_cap_names(geo)[list(caps.keys())[np.argmax(err[-1])]]
            f_delta = (f_conv - norm) * post.convert[f]
            out_str = geo + ' ' + f[:4]
            out_str += '\tdelta to periodic ' + '{:2.1e}'.format(f_delta) + ' [' + post.units[f] + '] '
            out_str += '\tperiodic error {:.2e}'.format(max_err * 100) + '%'
            out_str += '\tconverged in {:>2}'.format(str(i_conv)) + ' cycles'
            if f == 'pressure':
                out_str += ' ' + '\t{:.2e}'.format(i_conv / np.max(tau_all) * dt) + ' times tau'
            else:
                out_str += '\t\t\t'
            out_str += '\tat outlet ' + max_outlet
            print(out_str)

            # plot data points
            ax[pos].plot(np.arange(2, len(err) + 2), err, 'o-')

            # plot threshold
            ax[pos].plot([0, 999], [thresh, thresh], 'k-')

            # set plot options
            if i == 0:
                ax[pos].set_title(geo)
            if i == len(fields) - 1:
                ax[pos].set_xlabel('Cardiac cycle')
            if j == 0:
                ax[pos].set_ylabel(f.capitalize() + ' cyclic error')
            ax[pos].grid(True)
            ax[pos].ticklabel_format(axis='y')
            ax[pos].set_yscale('log')
            ax[pos].set_ylim([1.0e-5, 1])
            # ax[pos].set_xlim([2, np.max(cycles)])
            ax[pos].set_xlim([2, 50])
            ax[pos].yaxis.set_major_formatter(mtick.PercentFormatter(1.0, 3))

    plt.subplots_adjust(right=0.8)
    fname = 'outlets.png'
    fpath = os.path.join(db.get_statistics_dir(), fname)
    fig.savefig(fpath, bbox_inches='tight')
    plt.close(fig)


def main(db, geometries, params):
    geos = []
    for geo in geometries:
        f_path = db.get_3d_flow_rerun(geo)
        if not os.path.exists(f_path):
            continue
        geos += [geo]

    # geos = ['0003_0001', '0067_0001']
    plot_error_spatial(db, geos)


if __name__ == '__main__':
    descr = 'Get 3D-3D statistics'
    d, g, p = input_args(descr)
    main(d, g, p)



#!/usr/bin/env python

import numpy as np
import os
import glob
import pdb
import vtk

from get_database import Database, input_args
from vtk_functions import read_geo, write_geo


def invert_traingles(surf, name):
    """
    Invert triangles by flipping nodes from  to
    """
    cells = vtk.vtkCellArray()
    for i in range(surf.GetNumberOfCells()):
        assert surf.GetCellType(i) == 5, 'not a triangle mesh'

        # could be done easier with DeepCopy but throws seg fault
        ids = surf.GetCell(i).GetPointIds()
        id0 = ids.GetId(0)
        id1 = ids.GetId(1)
        id2 = ids.GetId(2)

        tri = vtk.vtkTriangle()
        if 'all_exterior' in name or 'wall' in name:
            tri.GetPointIds().SetId(0, id2)
            tri.GetPointIds().SetId(1, id1)
            tri.GetPointIds().SetId(2, id0)
        else:
            tri.GetPointIds().SetId(0, id1)
            tri.GetPointIds().SetId(1, id2)
            tri.GetPointIds().SetId(2, id0)
        cells.InsertNextCell(tri)
    surf.SetPolys(cells)
    surf.Modified()


def fix_surf_orientation(db, geo):
    # get all surfaces
    surfaces = glob.glob(os.path.join(db.get_surface_dir(geo), '*.vtp'))

    for f_surf in surfaces:
        name = os.path.basename(f_surf)
        # if name == 'btrunk.vtp' or name == 'carotid.vtp':
        print('  fixing ' + name)

        surf = read_geo(f_surf).GetOutput()
        invert_traingles(surf, name)

        f_out = os.path.join(db.fpath_gen, 'surfaces', geo + '_flipped', name)
        write_geo(f_out, surf)


def main(db, geometries):
    for geo in geometries:
        print('Fixing geometry ' + geo)

        fix_surf_orientation(db, geo)


if __name__ == '__main__':
    descr = 'Fix orientation of surface triangles'
    d, g, _ = input_args(descr)
    main(d, g)
